/*
           *** cmdftp, my ftp client, sick_soul 2k3 ***

    Copyright (C) 2003 Claudio Fontana

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (look for the file called COPYING);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    You can contact the author (Claudio Fontana) by sending a mail
    to sick_soul@users.sourceforge.net
	                                                                 */
#include "cmdftp.h"

/*******************************/
/* ERROR, WARNING, ... STRINGS */
/*******************************/

char* msg_err[] = {
  "bye", "bad server response", "unrecognized option:", "wrong option value:", "option without value:", "connection failed", "access denied", "unexpected error", "heap allocation failed"
};

char* msg_war[] = {
  "cannot open local file:", "error writing to local file:", "error reading from local file:", "transfer not confirmed", "no file matching:", "cannot open temp file:", "unknown command:", "missing argument", "login failed", "operation failed or incomplete:", "missing environment variable:", "Resuming connection lost during", "interrupted during"
};

char* msg_usage = "cmdftp [option [option_value]]* hostname\n\n"
"options: \n"
" -p FTP_PORT    specify the remote TCP port of the server (def:21)\n"
" -b BUFFER_SIZE specify the size in bytes of the transfer buffer (def:8192)\n"
" -q             be quiet\n"
" -n             disable autologin (~/.netrc)\n"
" -g             disable output paging\n"
" -d             disable forced ignore of dot-files\n";

char* msg_intro = "cmdftp version 0.7.3 Copyright (C) 2003 Claudio Fontana\n\n"
"cmdftp comes with ABSOLUTELY NO WARRANTY; this is free software, and you\n"
"are welcome to redistribute it under certain conditions specified by the\n"
"GPL (GNU General Public License). Look at the COPYING file for the details.\n"
"\nType h after the login to get help about internal commands.\n";

char* msg_help[] = {
  "h", "display this help",
  "l", "switch to local mode, following commands refer to local",
  "r", "switch to remote mode, following commands refer to remote",
  "cd PATH", "change working directory",
  "md PATH", "make directory",
  "rd PATH", "remove empty directory",
  "rm MASK", "delete regular files matching MASK, skip directories",
  "ls", "list current directory contents",
  "ls MASK", "list files/dirs matching MASK",
  "cp MASK DIR", "copy all files/dirs matching MASK to DIR (recurs)",
  "mv MASK DIR", "move all files/dirs matching MASK to DIR (recurs)",
  "u  MASK DIR", "upload files/dirs matching MASK into remote DIR (recurs)",
  "d  MASK DIR", "download files/dirs matching MASK into local DIR (recurs)",
  "dr MASK DIR", "same as above. If local file already exists, resume",
  "p  FILE", "print contents of the FILE on the terminal",
  "e  FILE", "edit FILE",
  "quit|exit|bye", "quit client",
  "<TAB>", "tab-completion"
};


/***********/
/* GLOBALS */
/***********/


struct cmdftp_options o = { 0, 21, 8192, 0, 0, 0, 0 };/* program options */

int cmdftp_control = 0;                               /* control connection */
int cmdftp_data = 0;                                  /* data connection */

char* user = 0; char* pass = 0;
int logging_in = 0;                                   /* user_logging_in? */
volatile int transfer_interrupted = TRAN_INTR_NO;     /* Why? */

struct hostent* server = 0;                           /* the remote host */
struct termios cmdftp_t;                              /* terminal data */

char* buffer;                                         /* in/out file transf */

char cmd_buffer[CMD_BUF_SIZE + 1];                    /* in/out commands */
char cmd_line[CMD_BUF_SIZE + 1];                      /* in     line sep */
char cmd_userinput[CMD_BUF_SIZE + 1];                 /* buf for term in */
char* cmd_ptr;                                        /* ->cmd_buffer */

char* env[N_ENV];                                     /* used env vars */
char* cwd[2] = { 0 };                                 /* working dir {l, r} */

unsigned char mode = CMDFTP_REMOTE;                   /* current mode {l, r} */

struct sigaction sa;

/*********************************************/
/* USER COMMANDS and corresponding functions */
/*********************************************/

char* commands[] = { "bye", "cd", "cp", "d", "dr", "e", "exit", "h", "l", "ls", "md", "mv", "p", "quit", "r", "rd", "rm", "u" };

#define N_COMMANDS (sizeof(commands) / sizeof(char*))

fun_ptr functions[2][N_COMMANDS] = {
  { &quit, &local_cd, &local_cp, &d, &dr, &local_e, &quit, &h, &local_mode, &local_ls, &local_md, &local_mv, &local_p, &quit, &remote_mode, &local_rd, &local_rm, &u },
  { &quit, &remote_cd, &remote_cp, &d, &dr, &remote_e, &quit, &h, &local_mode, &remote_ls, &remote_md, &remote_mv, &remote_p, &quit, &remote_mode, &remote_rd, &remote_rm, &u }
};


/*********************************/
/* ERROR, WARNING, ... FUNCTIONS */
/*********************************/


void cmdftp_err(int code, char* msg)
  { fprintf(stderr, "cmdftp: %s %s\n", msg_err[code], msg); exit(code); }

void cmdftp_war(int code, char* msg)
  { if (!o.q) fprintf(stderr, "cmdftp: %s %s\n", msg_war[code], msg); }

void usage() { printf("\n%s\n", msg_usage); exit(-1); }

void intro() {
  int i; if (o.q) return;
  for (i = 0; i < 79; i++) putchar('*');
  printf("\n%s\n", msg_intro); for (i = 0; i < 79; i++) putchar('*');
  putchar('\n');
}


/***************************************/
/* MAIN and other high level functions */
/***************************************/


int main(int argc, char* argv[]) {
  int attempt;

  if (argc < 2) usage();
  getoptions(argc, argv);

  intro();
  reset_cmd_buffer(); buffer = my_malloc(o.b);

  tcgetattr(0, &cmdftp_t); atexit(&cmdftp_canon_mode);
  init_signals();

  if (!(cmdftp_control = cmdftp_connect(o.p)))
    cmdftp_err(CMDFTP_ERR_CONN, "");

  greeting();

  for (attempt = 0; attempt < 3; attempt++) {
    user = pass = 0;
    if (!o.n && auto_login()) break;
    if (user) free(user); if (pass) free(pass);
    if (manual_login()) break;
    cmdftp_war(CMDFTP_WAR_LGIN, "");
    if (user) free(user); if (pass) free(pass);
  }

  if (attempt == 3) cmdftp_err(CMDFTP_ERR_LGIN, "");

  if ((cwd[CMDFTP_REMOTE] = remote_pwd()))
    cwd[CMDFTP_REMOTE] = strdup(cwd[CMDFTP_REMOTE]);
  else cmdftp_err(CMDFTP_ERR_SERV, "(PWD)");

  cwd[CMDFTP_LOCAL] = strdup(local_pwd());

  while (1) {
    char* cmd;
    printf("cmdftp:%s>", cwd[mode]); fflush(stdout);
    if (!(dispatch(cmd = readline(1, 1)))) return 0; 
    free(cmd);
  }
}

/*****************************************************************************/

void getoptions(int argc, char* argv[]) {
  int i;

  for (i = 1; i < argc; i++) {
    if (i == argc - 1)
      if (argv[i][0] != '-') o.hostname = strdup(argv[i]);
      else usage();
    else if (argv[i][0] == '-') {
      int rv = 0;
      switch(argv[i][1]) {
      case 'p' : rv = sscanf(argv[++i], "%hu", &o.p); break;
      case 'b' : rv = sscanf(argv[++i], "%hu", &o.b); break;
      case 'h' : usage();
      case 'q' : o.q = 1; rv = 1; break;
      case 'n' : o.n = 1; rv = 1; break;
      case 'g' : o.g = 1; rv = 1; break;
      case 'd' : o.d = 1; rv = 1; break;
      default: cmdftp_err(CMDFTP_ERR_UOPT, argv[i]);
      }
      if (rv != 1) cmdftp_err(CMDFTP_ERR_WOPT, argv[i]);
    } else cmdftp_err(CMDFTP_ERR_NOPT, argv[i]);
  }

  env[CMDFTP_ENV_PAGER] = o.g ? 0 : getenv("PAGER");
  env[CMDFTP_ENV_EDITOR] = getenv("EDITOR");
  env[CMDFTP_ENV_HOME] = getenv("HOME");
}

/*****************************************************************************/

void greeting() {
  if (recv_answer(0, 0) != 220) cmdftp_err(CMDFTP_ERR_SERV, "(greeting)");
}

/*****************************************************************************/

int login(char* user, char* pass) {
  int rv = 0;

  logging_in = 1;
  sprintf(cmd_buffer, "USER %s", user);
  if (send_command(cmd_buffer) != 331) return 0;

  sprintf(cmd_buffer, "PASS %s", pass);
  if (send_command(cmd_buffer) == 230) rv = 1;
  logging_in = 0;
  return rv;
}

int manual_login() {
  printf("Username: ");
  if (!(user = readline(0, 1))) return 0;

  cmdftp_pwd_start(); printf("Password: ");
  pass = readline(0, 0);
  cmdftp_pwd_end();

  if (!pass) return 0;
  return login(user, pass);
}

int auto_login() {
#define AUTO_LINE_LEN 80
  FILE* f; char* fname; char line[AUTO_LINE_LEN]; char token[AUTO_LINE_LEN];
  int state, rv;
  char* rc = ".netrc"; state = rv = 0;

  if (!env[CMDFTP_ENV_HOME]) return 0;
  fname = my_malloc(strlen(env[CMDFTP_ENV_HOME]) + 1 + strlen(rc) + 1);
  sprintf(fname, "%s/%s", env[CMDFTP_ENV_HOME], rc);
  f = fopen(fname, "r"); free(fname); if (!f) return 0;

  while((fgets(line, AUTO_LINE_LEN, f))) {
    char* ptr, *ptr_new, *ptr_tmp; int l;
    l = strlen(line); if (line[l - 1] == '\n') line[l - 1] = 0;
    if ((ptr = strchr(line, '#')))
      if (ptr == line || ptr[-1] != '\\') *ptr = 0;
    ptr = ptr_new = ptr_tmp = line;

  tokenize:
    if ((l = auto_login_next_token(&ptr, token)) == -1) continue;
    if (l == 0) {
      if (!state)
	{ if (strcmp(o.hostname, token) == 0) state++; goto tokenize; }
      break;
    } else if (l == 1) {
      if (!state) { state++; goto tokenize; }
      break;
    } else if (l == 2) {
      if (state && !user) user = strdup(token); goto tokenize;
    } else if (l == 3) {
      if (state && !pass) pass = strdup(token); goto tokenize;
    }
  } 
  if (user && !pass) pass = strdup("");
  if (user) rv = login(user, pass);
  if (!rv) {
    if (user) { free(user); user = 0; }
    if (pass) { free(pass); pass = 0; }
  }
  return rv;
#undef AUTO_LINE_LEN
}

int auto_login_next_token(char** hay, char* store) {
#define AUTO_N_KEYS 4
  char* keys[] = { "machine", "default", "login", "password" };
  int keys_len[] = { 7, 7, 5, 8 };
  char* ptr, *minptr; int i, min;

  min = -1; minptr = 0;

  for (i = 0; i < AUTO_N_KEYS; i++)
    if ((ptr = strstr(*hay, keys[i])))
      if (!minptr || ptr < minptr) { minptr = ptr; min = i; }

  if (min != -1) {
    *hay = minptr + keys_len[min];
    if (min != 1) {
      if (sscanf(*hay, " %s%n", store, &i) == 0) { i = 0; *store = 0; }
      *hay += i;
    }
  }
  return min;
#undef AUTO_N_KEYS
}


/***********************/
/* FUNCTION DISPATCHER */
/***********************/


int dispatch(char* cmd) {
  int idx;
  char* argv[2];

  if (!cmd || !*cmd) return 0;
  split_cmd(cmd, argv);

  if ((idx = str_binsearch(argv[0])) == -1)
    cmdftp_war(CMDFTP_WAR_BADC, cmd);
  else if (!functions[mode][idx](argv[1]))
    cmdftp_war(CMDFTP_WAR_FAIL, cmd);
  return 1;
}


/*********************************/
/* DISPATCHED FUNCTIONS: general */
/*********************************/


int quit(char* arg) { cmdftp_err(CMDFTP_ERR_NULL, ""); return 1; }

int h(char* arg) { 
  int i;
  
  for (i = 0; i < (sizeof(msg_help) / sizeof(char*)); i += 2)
    printf("%-14s %s\n", msg_help[i], msg_help[i + 1]);
  return 1;
}
    
int local_mode(char* arg) {
  mode = CMDFTP_LOCAL; printf("Current mode: local file system\n"); return 1;
}

int remote_mode(char* arg) {
  mode = CMDFTP_REMOTE; printf("Current mode: remote file system\n"); return 1;
}


/************************************/
/* DISPATCHED FUNCTIONS: local mode */
/************************************/


int local_cd(char* arg) {
  int rv = 0;
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (local_cd_aux(arg)) rv = 1;
  free(cwd[CMDFTP_LOCAL]); cwd[CMDFTP_LOCAL] = strdup(local_pwd());
  return rv;
}

int local_cd_aux(char* arg) { return (chdir(arg) == 0); }

/*****************************************************************************/

int local_cp(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if ((!local_cd_aux(paths[1])) && (local_cd_aux(paths[0]))) {
    source_mask = fullpath(paths[0], "*");
  } else source_mask = strdup(paths[0]);

  local_cd_aux(cwd[CMDFTP_LOCAL]);
  rv = local_cp_aux(paths[1], source_mask);
  free(source_mask);
  return rv;
}

/*****************************************************************************/

int local_cp_aux(char* target_dir, char* source_mask) {
  struct list_data d;
  char* target[3];
  int i, rv = 1;

  if (!local_cd_aux(target_dir)) local_md(target_dir);
  else local_cd_aux(cwd[CMDFTP_LOCAL]);

  init_list(&d);

  if (!local_fetch_list(source_mask, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  canonized_fn(target, target_dir);

  for (i = 0; i < d.count; i++) {
    struct stat si;
    if (stat(d.data[i].fullname, &si) != -1) {

      if (S_ISREG(si.st_mode)) {
	if (!local_copy(target[2], d.data + i)) { rv = 0; break; }

      } else if (S_ISDIR(si.st_mode)) {
	char* tmp_src, *tmp_trg; int subrv = 0;

	tmp_trg = fullpath(target[2], d.data[i].basename);
	tmp_src = fullpath(d.data[i].fullname, "*");
	subrv = local_cp_aux(tmp_trg, tmp_src);
	free(tmp_trg); free(tmp_src);
	if (!subrv) { rv = 0; break; }
      }
    }
  }
  free_list(&d); free_fn(target); 
  return rv;
}

/*****************************************************************************/

int local_mv(char* arg) {
  char* paths[2]; struct list_data d;
  char* target[3];
  int i, rv = 1;

  split_cmd(arg, paths);

  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }

  init_list(&d);
  if (!local_fetch_list(paths[0], &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, paths[0]); return 1; }

  canonized_fn(target, paths[1]);

  for (i = 0; i < d.count; i++) {
    struct stat si;

    if (stat(d.data[i].fullname, &si) != -1) {
      if (S_ISREG(si.st_mode) || S_ISDIR(si.st_mode)) {
	char* fn;
	if (!local_md(target[2])) fn = fullpath(target[2], d.data[i].basename);
	else fn = strdup(target[2]);
	if (rename(d.data[i].fullname, fn) == -1) { free(fn); rv = 0; break; }
	free(fn);
      }
    }
  }
  free_list(&d); free_fn(target); 
  return rv;
}

/*****************************************************************************/

int local_copy(char* target_dir, struct list_entry* source) {
  FILE* t, *s;
  char* target;
  int c, rv = 0;
  
  if (!(s = fopen(source->fullname, "rb"))) return 0;
  target = fullpath(target_dir, source->basename);
  
  if ((t = fopen(target, "wb"))) {
    while ((c = getc(s)) != EOF) putc(c, t);
    rv = 1;
  }

  free(target); if (t) fclose(t); fclose(s); return rv;
}

/*****************************************************************************/

int u(char* arg) {
  char* paths[2], *source_mask;
  int rv = 0;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!remote_cd_aux(paths[1])) && (local_cd_aux(paths[0]))) {
    local_cd_aux(cwd[CMDFTP_LOCAL]);
    source_mask = fullpath(paths[0], "*");
  } else source_mask = strdup(paths[0]);

  remote_cd_aux(cwd[CMDFTP_REMOTE]);
  rv = u_aux(paths[1], source_mask);
  free(source_mask);
  return rv;
}

/*****************************************************************************/

int u_aux(char* target_dir, char* source_mask) {
  struct list_data ld;
  char* target[3];
  int i, rv = 1;

  if (!remote_cd_aux(target_dir)) remote_md(target_dir);
  else remote_cd_aux(cwd[CMDFTP_REMOTE]);
  
  init_list(&ld);

  if (!local_fetch_list(source_mask, &ld)) return 0;
  if (ld.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  canonized_fn(target, target_dir);
  
  for (i = 0; i < ld.count; i++) {
    struct stat si;
    if (stat(ld.data[i].fullname, &si) != -1) {
      if (S_ISREG(si.st_mode)) {
	FILE* f;

	if (!(f = fopen(ld.data[i].fullname, "rb")))
	  cmdftp_war(CMDFTP_WAR_OPEN, ld.data[i].fullname);
	else {
	  char* fn; int up;
	  fn = fullpath(target[2], ld.data[i].basename);
	  up = upload(fn, f);
	  fclose(f); free(fn);
	  if (!up) { rv = 0; break; }
	}
      } else if (S_ISDIR(si.st_mode)) {
	char* tmp_src, *tmp_trg; int subrv = 0;

	tmp_src = fullpath(ld.data[i].fullname, "*");
	tmp_trg = fullpath(target[2], ld.data[i].basename);

	subrv = u_aux(tmp_trg, tmp_src);
	free(tmp_trg); free(tmp_src);
	if (!subrv) { rv = 0; break; }
      }
    }
  }
  free_list(&ld); free_fn(target);
  return rv;
}

/*****************************************************************************/

int upload(char* target, FILE* source) {
  int start_pos, total_size, rv = 0;
  int port;
  int bytes = 0; int written = 0;

  if (!send_command("TYPE I")) return 0;

 start_transfer:

  fseek(source, 0, SEEK_END); total_size = ftell(source); fseek(source, 0, SEEK_SET);
  start_pos = 0;

  if (!(port = getport())) return 0;
  if (!(cmdftp_data = cmdftp_connect(port))) return 0;

  sprintf(cmd_buffer, "STOR %s", target);
  if (send_command(cmd_buffer) != 150) { close(cmdftp_data); return 0; }
  
  printf("Uploading %s , %i bytes.\n", clean_fn(target), total_size);

  transfer_interrupted = TRAN_INTR_NO;

  while (start_pos < total_size) {
    printf("%3i%%\r", (start_pos * 100 / total_size)); fflush(stdout);
    if ((bytes = fread(buffer, 1, min(o.b, total_size - start_pos), source)) <= 0 || (written = my_raw_write(buffer, bytes, cmdftp_data)) != bytes) break;

    start_pos += bytes;
  }

  close(cmdftp_data); 

  if (start_pos < total_size) {
    if (bytes < 0) cmdftp_war(CMDFTP_WAR_READ, basename(target));
    else TRANSFER_INTERRUPTED_CHECK("data transfer", 1);
    else { cmdftp_err(CMDFTP_ERR_UNER, strerror(errno)); }
  } else {
    if (!recv_confirm()) cmdftp_war(CMDFTP_WAR_NOCT, "");
    rv = 1;
  }

  return rv;
}

/*****************************************************************************/

int local_ls(char* arg) {
  struct list_data d;
  int rv;

  init_list(&d);
  if (!local_fetch_list(arg, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  rv = ls(&d); free_list(&d); return rv;
}

/*****************************************************************************/

int local_md(char* arg) {
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (mkdir(arg, 0777) < 0) return 0;
  return 1;
}

/*****************************************************************************/

int local_rd(char* arg) { 
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (rmdir(arg) < 0) return 0;
  return 1;
}

/*****************************************************************************/

int local_rm(char* arg) {
  struct list_data d;
  int i, rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }

  init_list(&d);
  if (!local_fetch_list(arg, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  for (i = 0; i < d.count; i++) {
    struct stat si;
    if ((stat(d.data[i].fullname, &si) != -1) && (S_ISREG(si.st_mode)))
      if (unlink(d.data[i].fullname) == -1) break;
    if (i == d.count - 1) rv = 1;
  }
  free_list(&d);
  return rv;
}

/*****************************************************************************/

int local_p(char* arg) {

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }

  if (env[CMDFTP_ENV_PAGER])
    return cmdftp_execute(env[CMDFTP_ENV_PAGER], arg, 0);
  else {
    FILE* f; int c;
    if (!(f = fopen(arg, "r"))) return 0;
    while ((c = getc(f)) != EOF) putchar(c); printf("\n");
    fclose(f);
  }
  return 1;
}

/*****************************************************************************/

int local_e(char* arg) {
  
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (!env[CMDFTP_ENV_EDITOR])
    { cmdftp_war(CMDFTP_WAR_MENV, "EDITOR"); return 0; }

  return cmdftp_execute(env[CMDFTP_ENV_EDITOR], arg, 0);
}


/*************************************/
/* DISPATCHED FUNCTIONS: remote mode */
/*************************************/


int remote_cd(char* arg) {
  int rv = 0;
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (remote_cd_aux(arg)) rv = 1;

  free(cwd[CMDFTP_REMOTE]);
  
  if ((cwd[CMDFTP_REMOTE] = remote_pwd()))
    cwd[CMDFTP_REMOTE] = strdup(cwd[CMDFTP_REMOTE]);
  else if ((cwd[CMDFTP_REMOTE] = remote_pwd()))
    cwd[CMDFTP_REMOTE] = strdup(cwd[CMDFTP_REMOTE]);
  else cmdftp_err(CMDFTP_ERR_SERV, "(PWD)");
				
  return rv;
}

int remote_cd_aux(char* arg) {
  sprintf(cmd_buffer, "CWD %s", arg);
  return (send_command(cmd_buffer) == 250);
}

/*****************************************************************************/

int d(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!local_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    remote_cd_aux(cwd[CMDFTP_REMOTE]);
    source_mask = fullpath(paths[0], "*");
  } else source_mask = strdup(paths[0]);

  local_cd_aux(cwd[CMDFTP_LOCAL]);
  rv = d_aux(paths[1], source_mask, 0);
  free(source_mask);
  return rv;
}

/*****************************************************************************/

int dr(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!local_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    remote_cd_aux(cwd[CMDFTP_REMOTE]);
    source_mask = fullpath(paths[0], "*");
  } else source_mask = strdup(paths[0]);

  local_cd_aux(cwd[CMDFTP_LOCAL]);
  rv = d_aux(paths[1], source_mask, 1);
  free(source_mask);
  return rv;
}

/*****************************************************************************/

int d_aux(char* target_dir, char* source_mask, int resume) {
  struct list_data ld;
  char* target[3];
  int i, rv = 1;

  if (!local_cd_aux(target_dir)) local_md(target_dir);
  local_cd_aux(cwd[CMDFTP_LOCAL]);

  init_list(&ld);

  if (!remote_fetch_list(source_mask, &ld)) return 0;
  if (ld.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  canonized_fn(target, target_dir);
  
  for (i = 0; i < ld.count; i++) {
    if (!remote_cd_aux(ld.data[i].fullname)) {
      char* fn; FILE* f;
      
      fn = fullpath(target[2], ld.data[i].basename);
      if (!(f = fopen(fn, resume ? "ab" : "wb")))
	cmdftp_war(CMDFTP_WAR_OPEN, fn);
      else {
	int dl = download(f, ld.data[i].fullname);
	fclose(f);
	if (!dl) { rv = 0; free(fn); break; }
      }
      free(fn);
    } else {
      char* tmp_src, *tmp_trg; int subrv = 0;

      remote_cd_aux(cwd[CMDFTP_REMOTE]); 
      tmp_src = fullpath(ld.data[i].fullname, "*");
      tmp_trg = fullpath(target[2], ld.data[i].basename);
      subrv = d_aux(tmp_trg, tmp_src, resume);
      free(tmp_trg); free(tmp_src);
      if (!subrv) { rv = 0; break; }
    }
  }

  free_list(&ld); free_fn(target);
  return rv;
}

/*****************************************************************************/

int remote_cp(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!remote_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    source_mask = fullpath(paths[0], "*");
  } else source_mask = strdup(paths[0]);

  remote_cd_aux(cwd[CMDFTP_REMOTE]);
  rv = remote_cp_aux(paths[1], source_mask);
  free(source_mask);
  return rv;
}

/*****************************************************************************/

int remote_cp_aux(char* target_dir, char* source_mask) {
  struct list_data d;
  char* target[3];
  int i, rv = 1;

  init_list(&d);

  if (!remote_fetch_list(source_mask, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  if (!remote_cd_aux(target_dir)) remote_md(target_dir);
  else remote_cd_aux(cwd[CMDFTP_REMOTE]);

  canonized_fn(target, target_dir);
  
  for (i = 0; i < d.count; i++) {
    if (!remote_cd_aux(d.data[i].fullname)) {
      if (!remote_copy(target[2], d.data + i)) { rv = 0; break; }
    } else {
      char* tmp_src, *tmp_trg; int subrv = 0;
      
      remote_cd_aux(cwd[CMDFTP_REMOTE]);
      tmp_src = fullpath(d.data[i].fullname, "*");
      tmp_trg = fullpath(target[2], d.data[i].basename);
      subrv = remote_cp_aux(tmp_trg, tmp_src);
      free(tmp_src); free(tmp_trg);
      if (!subrv) { rv = 0; break; }
    }
  }

  free_list(&d); free_fn(target);
  return rv;
}

/*****************************************************************************/

int remote_mv(char* arg) {
  char* paths[2], *source_mask;
  int rv, rd = 0;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!remote_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    rd = 1;
    source_mask = fullpath(paths[0], "*");
  } else source_mask = strdup(paths[0]);

  remote_cd_aux(cwd[CMDFTP_REMOTE]);
  rv = remote_mv_aux(paths[1], source_mask);
  if (rd) remote_rd(paths[0]);
  free(source_mask);
  return rv;
}

/*****************************************************************************/

int remote_mv_aux(char* target_dir, char* source_mask) {
  struct list_data d;
  char* target[3];
  int i, rv = 1;

  init_list(&d);

  if (!remote_fetch_list(source_mask, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  if (!remote_cd_aux(target_dir)) remote_md(target_dir);
  else remote_cd_aux(cwd[CMDFTP_REMOTE]);

  canonized_fn(target, target_dir);
  
  for (i = 0; i < d.count; i++) {
    if (!remote_cd_aux(d.data[i].fullname)) {
      sprintf(cmd_buffer, "RNFR %s", d.data[i].fullname);
      if (send_command(cmd_buffer) != 350)
	{ cmdftp_war(CMDFTP_WAR_MASK, d.data[i].fullname); goto end_proc; }
      
      sprintf(cmd_buffer, "RNTO %s/%s", target[2], d.data[i].basename);
      if (send_command(cmd_buffer) != 250) { rv = 0; goto end_proc; }
    } else {
      char* tmp_src, *tmp_trg; int subrv = 0;
      
      remote_cd_aux(cwd[CMDFTP_REMOTE]);
      tmp_src = fullpath(d.data[i].fullname, "*");
      tmp_trg = fullpath(target[2], d.data[i].basename);
      subrv = remote_mv_aux(tmp_trg, tmp_src);
      free(tmp_src); free(tmp_trg);
      if (subrv) remote_rd(d.data[i].fullname);
      else { rv = 0; break; }
    }
  }

 end_proc:

  free_list(&d); free_fn(target);
  return rv;
}

/*****************************************************************************/

int remote_copy(char* target_dir, struct list_entry* source) {
  
  FILE* t;
  char* t_fn, *target = 0;
  int len, rv = 0;

  if (!(t = cmdftp_temp(&t_fn)))
    { cmdftp_war(CMDFTP_WAR_TEMP, t_fn); return 0; }

  if (!download(t, source->fullname)) goto end_proc;

  len = strlen(target_dir);
  if (target_dir[len - 1] == '/') target_dir[(--len)] = 0;

  target = fullpath(target_dir, source->basename);
  if (upload(target, t)) rv = 1;

 end_proc:
  fclose(t); unlink(t_fn); free(t_fn);
  if (target) free(target); return rv;
}

/*****************************************************************************/

int download(FILE* target, char* source) {
  int start_pos, total_size, answer, rv = 0;
  int port;
  int bytes = 0; int written = 0;
  
  if (!send_command("TYPE I")) return 0;
  if ((total_size = getsize(source)) < 0) return 0;

 start_transfer:

  start_pos = ftell(target);

  sprintf(cmd_buffer, "REST %i", start_pos);
  if (!(answer = send_command(cmd_buffer))) return 0;

  if (start_pos > 0 && answer != 350) {
    fseek(target, 0, SEEK_SET); start_pos = 0;
  }

  if (!(port = getport())) return 0;
  if (!(cmdftp_data = cmdftp_connect(port))) return 0;

  sprintf(cmd_buffer, "RETR %s", source);
  if (send_command(cmd_buffer) != 150) { close(cmdftp_data); return 0; }

  printf("Downloading %s, %i bytes", source, total_size);
  if (start_pos > 0) printf(", continuing at position %i", start_pos);
  printf(".\n");

  transfer_interrupted = TRAN_INTR_NO;

  while (start_pos < total_size) {
    printf("%3i%%\r", (start_pos * 100 / total_size)); fflush(stdout);
    if ((bytes = my_raw_read(buffer, min(o.b, total_size - start_pos), cmdftp_data)) <= 0 || (written = fwrite(buffer, bytes, 1, target)) != 1) break;
    start_pos += bytes;
  }

  close(cmdftp_data); 

  if (start_pos < total_size) {
    if (written < 0)
      cmdftp_war(CMDFTP_WAR_WRIT, basename(source));
    else TRANSFER_INTERRUPTED_CHECK("data transfer", 1);
    else { cmdftp_err(CMDFTP_ERR_UNER, strerror(errno)); }
  } else {
    if (!recv_confirm()) cmdftp_war(CMDFTP_WAR_NOCT, "");
    rv = 1;
  }

  return rv;
}

/*****************************************************************************/

int ls(struct list_data* d) {
  FILE* target; int pipe_fd[2];
  int i, rv = 0;

  if (env[CMDFTP_ENV_PAGER]) {
    if (pipe(pipe_fd) < 0) return 0;
    if (!(target = fdopen(pipe_fd[1], "w")))
      { close(pipe_fd[0]); close(pipe_fd[1]); return 0; }
  } else target = stdout;
  
  for (i = 0; i < d->count; i++)
    fprintf(target, "%s\n", d->data[i].fullname);

  if (!env[CMDFTP_ENV_PAGER]) return 1;

  fclose(target);
  rv = cmdftp_execute(env[CMDFTP_ENV_PAGER], 0, pipe_fd[0]);

  close(pipe_fd[0]);
  return rv;
}

/*****************************************************************************/

int remote_ls(char* arg) {
  struct list_data d;
  int rv;

  init_list(&d);
  if (!remote_fetch_list(arg, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  rv = ls(&d); free_list(&d); return rv;
}

/*****************************************************************************/

int remote_md(char* arg) {
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; } 
  sprintf(cmd_buffer, "MKD %s", arg); if (send_command(cmd_buffer) != 257) return 0;
  return 1;
}

/*****************************************************************************/

int remote_rd(char* arg) {
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; } 
  sprintf(cmd_buffer, "RMD %s", arg); if (send_command(cmd_buffer) != 250) return 0;
  return 1;
}

/*****************************************************************************/

int remote_rm(char* arg) {
  struct list_data d;
  int i, rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  init_list(&d);
  if (!remote_fetch_list(arg, &d)) goto end_proc;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  for (i = 0; i < d.count; i++) {
    sprintf(cmd_buffer, "DELE %s", d.data[i].fullname);
    if (!send_command(cmd_buffer)) goto end_proc;
  }
  rv = 1;
  
 end_proc:
  free_list(&d);
  return rv;
}

/*****************************************************************************/

int remote_p(char* arg) {
  int c; char* fn; FILE* f; int rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (!(f = cmdftp_temp(&fn))) { cmdftp_war(CMDFTP_WAR_TEMP, fn); return 0; }

  if (download(f, arg)) {
    if (env[CMDFTP_ENV_PAGER]) {
      fclose(f); f = 0;
      rv = cmdftp_execute(env[CMDFTP_ENV_PAGER], fn, 0);
    } else {
      fseek(f, 0, SEEK_SET);
      while ((c = getc(f)) != EOF) putchar(c); printf("\n");
      rv = 1;
    }
  }

  if (f) fclose(f); unlink(fn); free(fn); return rv;
}

/*****************************************************************************/

int remote_e(char* arg) {
  char* fn; FILE* f; int rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (!env[CMDFTP_ENV_EDITOR])
    { cmdftp_war(CMDFTP_WAR_MENV, "EDITOR"); return 0; }

  if (!(f = cmdftp_temp(&fn))) { cmdftp_war(CMDFTP_WAR_TEMP, fn); return 0; }

  if (download(f, arg)) {
    fclose(f); f = 0;
    if (cmdftp_execute(env[CMDFTP_ENV_EDITOR], fn, 0)) {
      if ((f = fopen(fn, "r")) && upload(arg, f)) rv = 1;
    }
  }
  if (f) fclose(f); unlink(fn); free(fn); return rv;
}


/******************************/
/* OTHER AUX FUNCTIONS: local */
/******************************/


char* local_pwd() { return getcwd(cmd_buffer, CMD_BUF_SIZE); }

int local_fetch_list(char* filemask, struct list_data* d) {
  char* fn_mask[3], *tmpmask;
  struct dirent* entry;
  DIR* dir;
  int rv = 0;
  
  if (*filemask == 0) tmpmask = strdup("*");
  else tmpmask = strdup(filemask);

  canonized_fn(fn_mask, tmpmask);

  if (!(dir = opendir(fn_mask[0]))) goto end_proc;

  while ((entry = readdir(dir))) {
    if ((fnmatch(fn_mask[1], entry->d_name, 0)) == 0) {
      char* full_name = fullpath(fn_mask[0], entry->d_name);
      store_list(clean_fn(full_name), d);
      free(full_name);
    }
  }

  closedir(dir); rv = 1;

 end_proc:
  free(tmpmask); free_fn(fn_mask);
  return rv;
}


/*******************************/
/* OTHER AUX FUNCTIONS: remote */
/*******************************/


char* remote_pwd() {
  char* start, *end;
  if (!send_command("PWD")) return 0;
  if ( ((end = strrchr(cmd_buffer, '"')) == 0) ||
       ((start = strchr(cmd_buffer, '"')) == 0) ) return 0;
  *end = 0; start++;
  return start;
}

/*****************************************************************************/

int remote_fetch_list(char* filemask, struct list_data* d) {
  char* tmpmask, *fn_mask[3], *line;
  int tmp, rv = 0;
  int port;

  if (!send_command("TYPE A")) return 0;

  if (*filemask == 0) tmpmask = strdup("*");
  else tmpmask = strdup(filemask);

  canonized_fn(fn_mask, tmpmask);

 start_transfer:
 
  if (!(port = getport())) goto end_proc;
  if (!(cmdftp_data = cmdftp_connect(port))) goto end_proc;

  sprintf(cmd_buffer, "NLST %s", fn_mask[0]);
  if ((tmp = (send_command(cmd_buffer))) != 150)
    { close(cmdftp_data); if (tmp == 450) rv = 1; goto end_proc; }

  reset_cmd_buffer(); 

  transfer_interrupted = TRAN_INTR_NO;

  while ((line = recv_line(cmdftp_data))) {

    char* fn = strrchr(line, '/');
    if (!fn) fn = line;
    else fn++;

    if (fnmatch(fn_mask[1], fn, 0) == 0) {
      char* full_name = fullpath(fn_mask[0], fn);
      store_list(clean_fn(full_name), d);
      free(full_name);
    }
  }

  close(cmdftp_data); 
  
  TRANSFER_INTERRUPTED_CHECK("data transfer", 1);
  else {
    if (!recv_confirm()) cmdftp_war(CMDFTP_WAR_NOCT, "");
    rv = 1;
  }
  
 end_proc:
  free(tmpmask); free_fn(fn_mask);
  return rv;
}  


/*****************************/
/* NETWORK RELATED FUNCTIONS */
/*****************************/


int cmdftp_connect(int port) {
  struct sockaddr_in address;
  int s; int opvalue = 8; socklen_t slen = sizeof(opvalue);

  memset(&address, 0, sizeof(address));

  if ( ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) ||
       ((setsockopt(s, IPPROTO_IP, IP_TOS, &opvalue, slen)) < 0) ) return 0;

  address.sin_family = AF_INET;
  address.sin_port = htons((unsigned short)port);

  if (!server)
    if (!(server = gethostbyname(o.hostname))) return 0;

  memcpy(&address.sin_addr.s_addr, server->h_addr, server->h_length);
  
  if (connect(s, (struct sockaddr*) &address, sizeof(address)) == -1)
    return 0;

  return s;
}

/*****************************************************************************/

void cmdftp_reconnect() {
  if (!(cmdftp_control = cmdftp_connect(o.p)))
    cmdftp_err(CMDFTP_ERR_CONN, "");
  greeting();
  if (!logging_in) {
    if (!login(user, pass)) cmdftp_err(CMDFTP_ERR_LGIN, "");
    if (cwd[CMDFTP_REMOTE])
      remote_cd(cwd[CMDFTP_REMOTE]);
  }
}

/*****************************************************************************/

int my_raw_read(char* buf, int n, int sc) { 
  int rv;

  rv = recv(sc, buf, n, 0);

  if (rv < 0) {
    if (transfer_interrupted != TRAN_INTR_INT)
      transfer_interrupted = TRAN_INTR_PIPE;
  }
  return rv;
}

/*****************************************************************************/

int my_raw_write(char* buf, int n, int sc) {
  int rv;

  rv = send(sc, buf, n, 0);

  if (rv < 0) {
    if (transfer_interrupted != TRAN_INTR_INT)
      transfer_interrupted = TRAN_INTR_PIPE;
  }
  return rv;
}

/*****************************************************************************/


int send_command(char* cmd) { 
  char* s; int len, rv = 0;

  s = my_malloc((len = strlen(cmd) + 2) + 1);
  strcpy(s, cmd); strcat(s, "\r\n");

 start_transfer:
  transfer_interrupted = TRAN_INTR_NO;

  if (my_raw_write(s, len, cmdftp_control) < 0) {
    s[len - 2] = 0;
    TRANSFER_INTERRUPTED_CHECK(s, 0);
  } else {
    s[len - 2] = 0;
    if (!(rv = recv_answer(0, 0))) {
      TRANSFER_INTERRUPTED_CHECK(s, 1);
    }
  }
  free(s);
  return rv;
}

/*****************************************************************************/

void reset_cmd_buffer() {
  cmd_ptr = cmd_buffer + CMD_BUF_SIZE;
}

/*****************************************************************************/

char* recv_line(int sc) {
  char* new_ptr;
  int n, len;

  if (*cmd_ptr == 0) {
    int bytes = my_raw_read(cmd_buffer, CMD_BUF_SIZE, sc);
    if (bytes <= 0) return 0;
    cmd_buffer[bytes] = 0;
    cmd_ptr = cmd_buffer;
    if (*cmd_ptr == '\n' && cmd_buffer[CMD_BUF_SIZE - 1] == '\r') cmd_ptr++;
  }

  new_ptr = strchr(cmd_ptr, '\r');
  if (!new_ptr) new_ptr = cmd_buffer + CMD_BUF_SIZE;

  n = new_ptr - cmd_ptr;
  strncpy(cmd_line, cmd_ptr, n); cmd_line[n] = 0;
  cmd_ptr = new_ptr;

  if (cmd_ptr == cmd_buffer + CMD_BUF_SIZE) {
    int bytes = my_raw_read(cmd_buffer, CMD_BUF_SIZE, sc);
    if (bytes <= 0) return 0;
    cmd_buffer[bytes] = 0;

    cmd_ptr = cmd_buffer;

    new_ptr = strchr(cmd_ptr, '\r');
    if (!new_ptr) new_ptr = cmd_buffer + CMD_BUF_SIZE;
        
    n = new_ptr - cmd_ptr; len = strlen(cmd_line);
    strncat(cmd_line, cmd_ptr, n); cmd_line[len + n] = 0;
    cmd_ptr = new_ptr;
  }

  if (*cmd_ptr == '\r') {
    cmd_ptr++;
    if (*cmd_ptr == '\n') cmd_ptr++;
  }

  return cmd_line;
}

/*****************************************************************************/

int recv_confirm(void) {
  int rv = 0; 
  sa.sa_handler = SIG_IGN; sigaction(SIGINT, &sa, 0);

  /* here the user can't interrupt. The problem is, if he breaks here
     he will lose sync with the server */

  if (recv_answer(0, 0) == 226) rv = 1;
  sa.sa_handler = &handler_INT; sigaction(SIGINT, &sa, 0);
  return rv;
}

int recv_answer(int store, struct line_data* d) {
  char* answer; int code; char str_code[] = { 0, 0, 0, ' ', 0 };

  reset_cmd_buffer();

  answer = recv_line(cmdftp_control);
  
  if ((!answer) || (strlen(answer) < 3)) return 0;

  strncpy(str_code, answer, 3);

  if (answer[3] == '-') {
    while ((answer = recv_line(cmdftp_control))) {       
      if (strncmp(answer, str_code, 4) == 0) break;
      if (store) store_line(answer, d);
      else printf("%s\n", answer);
    }
  }
  sscanf(str_code, "%i", &code);
  return code;
}

/*****************************************************************************/

int getport() {
  unsigned int b[6]; 
  int i, len, answer; char* port_str;

  if ((answer = send_command("PASV")) != 227) return 0;

  port_str = cmd_buffer + 4;
  len = strlen(port_str += 4);

  for (i = 0; i < len; i++) 
    if (!isdigit(port_str[i])) port_str[i] = ' ';
  
  if (sscanf(port_str, "%u %u %u %u %u %u", b, b + 1, b + 2, b + 3, b + 4, b + 5) != 6) return 0;

  return b[4] * 256 + b[5];
}

/*****************************************************************************/

int getsize(char* filename) {
  int size;

  sprintf(cmd_buffer, "SIZE %s", filename);
  if (send_command(cmd_buffer) != 213) { return -1; }

  sscanf(cmd_buffer + 4, "%i", &size);
  return size;
}



/*****************************/
/* VARIOUS UTILITY FUNCTIONS */
/*****************************/


void* my_malloc(size_t s) {
  void* rv;
  if (!(rv = malloc(s))) cmdftp_err(CMDFTP_ERR_HEAP, "");
  return rv;
}

void* my_realloc(void* ptr, size_t s) {
  void* rv;
  if (!(rv = realloc(ptr, s))) cmdftp_err(CMDFTP_ERR_HEAP, "");
  return rv;
}

/*****************************************************************************/

void split_cmd(char* cmd, char** argv) {
  if ((argv[1] = strchr(cmd, ' ')) == 0) argv[1] = cmd + strlen(cmd);
  else *argv[1]++ = 0;

  argv[0] = cmd;
}

/*****************************************************************************/

int str_binsearch(char* key) {
  int low, high, mid, chk;

  low = 0; high = N_COMMANDS - 1;

  while (low <= high) {
    mid = (low + high) / 2;
    if ((chk = strcmp(key, commands[mid])) > 0) low = mid + 1;
    else if (chk < 0) high = mid - 1;
    else return mid;
  }

  return -1;
}

/*****************************************************************************/

char* readline(int enable_tab, int enable_echo) {
  char* ptr; int c;

  ptr = cmd_userinput; memset(cmd_userinput, 0, sizeof(cmd_userinput));
  cmdftp_raw_mode();
  
  while (1) {
    c = fgetc(stdin);
    if (c == '\n') { *ptr = 0; fputc('\n', stdout); break; }
    else if ((c == '\t') && (enable_tab)) {
      readline_tab(); ptr = cmd_userinput + strlen(cmd_userinput);
    } else if (c == cmdftp_t.c_cc[VERASE]) readline_bs(&ptr, enable_echo);
    else if ((c >= 32) && (c <= 127)) {
      if ((ptr - cmd_userinput) < (CMD_BUF_SIZE - 10))
	{ *ptr++ = c; if (enable_echo) fputc(c, stdout); }
    }
    else if (c == EOF) break;
  }
  cmdftp_canon_mode();
  return (c != EOF) ? strdup(cmd_userinput) : 0;
}

/*****************************************************************************/

void readline_bs(char** ptr, int enable_echo) {
  if (*ptr > cmd_userinput) {
    *(--(*ptr)) = 0;
    if (enable_echo) fprintf(stdout, "\b \b");
  }
}

/*****************************************************************************/

void readline_tab() {
  char* cmd, *first, *second, *mask; struct list_data ld;
  int c, i, j, len, len_buffer, lc_mode;
  typedef int (*fetch_list_ptr)(char*, struct list_data*);
  fetch_list_ptr fetch_list[2] = { &local_fetch_list, &remote_fetch_list };

  cmd = first = second = mask = 0;
  c = i = j = len = len_buffer = 0;

  strcat(cmd_userinput, "*");
  init_list(&ld);
  if ((first = strchr(cmd_userinput, ' '))) {
    *first++ = 0;
    if ((second = strrchr(first, ' '))) {
      *second++ = 0; second = strdup(second);
    }
    first = strdup(first);
  } else { cmd = strdup(cmd_userinput); goto endtab; }
  
  cmd = strdup(cmd_userinput);
  
  if ((strcmp(cmd, "d") == 0) || (strcmp(cmd, "dr") == 0)) {
    lc_mode = second ? CMDFTP_LOCAL : CMDFTP_REMOTE;
  } else if (strcmp(cmd, "u") == 0) {
    lc_mode = second ? CMDFTP_REMOTE : CMDFTP_LOCAL;
  } else lc_mode = mode;
  
  if (!(mask = second)) mask = first;
  len_buffer = strlen(mask) - 1;
  
  if ( (!fetch_list[lc_mode](mask, &ld)) || (ld.count == 0) ) goto endtab;

  if ((len = strlen(ld.data[0].fullname)) + len_buffer + strlen(cmd) > CMD_BUF_SIZE - 10) {
    free_list(&ld); goto endtab;
  }
  
  for (i = 0; i < len; i++) {
    c = ld.data[0].fullname[i];
    for (j = 1; j < ld.count; j++) {
      if (ld.data[j].fullname[i] != c) goto endcommon;
    }
  }
  
 endcommon:
  ld.data[0].fullname[i] = 0;
 endtab:
  strcpy(cmd_userinput, cmd);
  if (first) {
    strcat(cmd_userinput, " ");
    if (second) { strcat(cmd_userinput, first); strcat(cmd_userinput, " "); }
  }
  if (ld.count) {
    strcat(cmd_userinput, ld.data[0].fullname);
    for (i = 0; i < len_buffer; i++) fputc('\b', stdout);
    fprintf(stdout, "%s", ld.data[0].fullname);
  } else {
    strncat(cmd_userinput, mask, len_buffer);
  }
  if (first) free(first); if (second) free(second); if (cmd) free(cmd);
  free_list(&ld);
}

/*****************************************************************************/

int min(int v1, int v2) { return v1 < v2 ? v1 : v2; }

/*****************************************************************************/

void cmdftp_pwd_start() {
  cmdftp_t.c_lflag &= ~ECHO; tcsetattr(0, TCSANOW, &cmdftp_t);
}

/*****************************************************************************/

void cmdftp_pwd_end() {
  cmdftp_t.c_lflag |= ECHO; tcsetattr(0, TCSANOW, &cmdftp_t);
}

/*****************************************************************************/

void cmdftp_raw_mode() {
  cmdftp_t.c_lflag &= ~ICANON; cmdftp_t.c_lflag &= ~ECHO;
  tcsetattr(0, TCSANOW, &cmdftp_t);
}

/*****************************************************************************/

void cmdftp_canon_mode() {
  cmdftp_t.c_lflag |= ICANON; cmdftp_t.c_lflag |= ECHO;
  tcsetattr(0, TCSANOW, &cmdftp_t);
}

/*****************************************************************************/

int cmdftp_execute(char* p1, char* p2, int pipe_fd) {
  char** argv = 0; char* argptr;
  int pid, status, rv = 0, count = 0;

  argptr = p1 = strdup(p1);

  do {
    argv = my_realloc(argv, sizeof(char*) * (count + 1));
    argv[count++] = argptr;
    argptr = strchr(argptr + 1, ' ');

    if (argptr) *argptr++ = 0;

  } while (argptr);

  count += (p2 ? 2 : 1); argv = my_realloc(argv, sizeof(char*) * count);
  argv[count - 1] = 0; if (p2) argv[count - 2] = p2;

  if ((pid = fork()) < 0) goto end_proc;
  else if (pid > 0) {
    if (waitpid(pid, &status, 0) > 0) {
      if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) rv = 1;
    }
  } else {
    if (!p2) { close(0); dup(pipe_fd); }
    execvp(argv[0], argv);
    exit(-1);
  }

 end_proc:
  free(p1); free(argv);
  return rv;
}

 
/*****************************************************************************/


char cmdftp_temp_fn[] = "/tmp/cmdftpXXXXXX";

FILE* cmdftp_temp(char** fn) {
  int i, d; char* ptr; FILE* rv;

  ptr = cmdftp_temp_fn + 11; *fn = 0;

  for (i = 0; i < 6; i++) ptr[i] = 'X';
  if ((d = mkstemp(cmdftp_temp_fn)) < 0) return 0;
  if (!(rv = fdopen(d, "r+"))) return 0;
  *fn = strdup(cmdftp_temp_fn);
  return rv;
}

/*****************************************************************************/

void canonized_fn(char* des[3], char* arg) {
  char* tmpmask;

  tmpmask = strdup(arg); des[0] = strdup(dirname(tmpmask)); free(tmpmask);
  tmpmask = strdup(arg); des[1] = strdup(basename(tmpmask)); free(tmpmask);
  des[2] = fullpath(des[0], des[1]);
}

/*****************************************************************************/

void free_fn(char* des[3]) {
  int i;
  for (i = 0; i < 3; i++) free(des[i]);
}

/*****************************************************************************/

char* clean_fn(char* fn) {
  if (strncmp(fn, "./", 2) == 0) return fn + 2;
  else if(strncmp(fn, "//", 2) == 0) return fn + 1;
  else return fn;
}

/*****************************************************************************/

char* fullpath(char* s, char* q) {
  char* rv = my_malloc(strlen(s) + strlen(q) + 2);
  sprintf(rv, "%s/%s", s, q);
  return rv;
}

/********************/
/* SIGNAL FUNCTIONS */
/********************/

void init_signals() {
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_IGN; sigaction(SIGPIPE, &sa, 0); 
  sa.sa_handler = &handler_INT; sigaction(SIGINT, &sa, 0);
}

void handler_INT(int i) { transfer_interrupted = TRAN_INTR_INT; }


/******************************/
/* STRUCT LINE_DATA FUNCTIONS */
/******************************/


void init_lines(struct line_data* d) {
  d->count = 0; d->lines = 0;
}

void init_list(struct list_data* d) {
  d->count = 0; d->data = 0;
}

void store_line(char* line, struct line_data* d) {
  int i = d->count;
  d->lines = my_realloc(d->lines, sizeof(char*) * (++d->count));
  d->lines[i] = my_malloc(CMD_BUF_SIZE - 9);
  strncpy(d->lines[i], line, CMD_BUF_SIZE - 10); d->lines[i][CMD_BUF_SIZE - 10] = 0; 
}

void store_list(char* fullname, struct list_data* d) {
  int i; char* base;
  base = basename(fullname);

  if ((o.d && (strcmp(base, ".") == 0 || strcmp(base, "..") == 0)) ||
      (!o.d && *base == '.')) return;
  
  i = d->count;
  d->data = my_realloc(d->data, sizeof(struct list_entry) * (++d->count));
  d->data[i].fullname = strdup(fullname);
  d->data[i].basename = strdup(base);
  d->data[i].dirname = strdup(dirname(fullname));
}

void free_lines(struct line_data* d) {
  int i;
  if (d->lines) {
    for (i = 0; i < d->count; i++) free(d->lines[i]);
    free(d->lines); d->lines = 0;
  } 
  d->count = 0;
}

void free_list(struct list_data* d) {
  int i;
  if (d->data) {
    for (i = 0; i < d->count; i++) {
      free(d->data[i].fullname);
      free(d->data[i].basename);
      free(d->data[i].dirname);
    }
    free(d->data); d->data = 0;
  }
  d->count = 0;
}


/* CMDFTP.C */
